#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import glob
import tempfile
from datetime import date, datetime
from pathlib import Path
from typing import Dict, List, Tuple

import xarray as xr
from dateutil.relativedelta import relativedelta
from loguru import logger

from weather_provider_api.routers.weather.repository.repository import (
    RepositoryUpdateResult,
)
from weather_provider_api.routers.weather.sources.cds.client import downloader

_FORMATTED_SUFFIX = ".FORMATTED.nc"
_UNFORMATTED_SUFFIX = ".UNFORMATTED.nc"
_INCOMPLETE_SUFFIX = ".INCOMPLETE.nc"
_TEMP_SUFFIX = ".TEMP.nc"


def _repository_callback(*args, **kwargs):  # pragma: no cover
    logger.info("Callback Received:")
    logger.info(" - Args: ", *args)
    logger.info(" - Kwargs: ", **kwargs)


def era5_update(
    file_prefix: str,
    repository_folder: str,
    repository_scope: Tuple[datetime, datetime],
    era5_dataset_name: str,
    era5_product_type: str,
    selected_factors: List[str],
    dataset_grid_resolution: Tuple[float, float],
    allowed_factors: Dict,
    maximum_allowed_runtime_in_minutes: int = (2 * 60),
    verify_last_available_date: bool = True,
):
    # Determine when to start and when to end
    start_of_update = datetime.utcnow()
    forced_end_of_update = start_of_update + relativedelta(minutes=maximum_allowed_runtime_in_minutes)

    # Set initial progress data
    current_months_processed = 0
    current_months_not_processable = 0
    average_time_per_month = 35  # Assuming 35 minutes processing time for the first item
    oldest_date_in_repo = repository_scope[0].date()
    newest_date_in_repo = repository_scope[1].date()

    logger.info(f"Started the update for repository [{file_prefix}]:")
    if verify_last_available_date:
        logger.info(f"Determining the most recent date available for [{era5_dataset_name}],[{era5_product_type}]")
        newest_date_in_repo = _get_actual_most_recent_data_for_model(
            newest_date_in_repo, era5_dataset_name, era5_product_type, dataset_grid_resolution
        )

    current_month_to_process = newest_date_in_repo.replace(day=1)
    while current_month_to_process >= oldest_date_in_repo.replace(day=1):
        if current_months_processed != 0 and current_months_processed > current_months_not_processable:
            average_time_per_month = (datetime.utcnow() - start_of_update).total_seconds() / current_months_processed

        if datetime.utcnow() + relativedelta(seconds=average_time_per_month) > forced_end_of_update:
            logger.info("There was not enough time left to process another file. Finishing up..")
            return RepositoryUpdateResult.timed_out

        try:
            process_month(
                file_prefix,
                repository_folder,
                current_month_to_process,
                era5_dataset_name,
                era5_product_type,
                dataset_grid_resolution,
                selected_factors,
                newest_date_in_repo,
                allowed_factors,
            )
        except Exception as e:
            logger.warning(
                f"Could not process month [{current_month_to_process.year}, {current_month_to_process.month}]: {e}"
            )
            current_months_not_processable += 1

        current_months_processed += 1
        current_month_to_process -= relativedelta(months=1)

    if current_months_processed > 1 and current_months_not_processable / current_months_processed > 0.3:
        # if over 30% of all requested months did not work and more than a month was requested return a failure value
        logger.error(
            "The margin of months that could not be processed exceeded set margins. Please contact an administrator."
        )
        return RepositoryUpdateResult.failure

    return RepositoryUpdateResult.completed


def process_month(
    file_name_prefix: str,
    repository_folder: str,
    month_to_process: date,
    dataset_name: str,
    product_name: str,
    grid_resolution: Tuple[float, float],
    factors: List[str],
    most_recent_data_date: date,
    allowed_factors: Dict,
):
    file_name = f"{file_name_prefix}_{month_to_process.year}_{str(month_to_process.month).zfill(2)}"
    file_path = Path(repository_folder).joinpath(file_name)

    if file_requires_update(file_path, month_to_process, most_recent_data_date):
        logger.debug(f"File update required for [{month_to_process.year}, {month_to_process.month}]")
        download_file = file_path.with_suffix(_UNFORMATTED_SUFFIX)
        download_era5_file(
            dataset=dataset_name,
            product_type=product_name,
            grid_resolution=grid_resolution,
            weather_factors=factors,
            years=[month_to_process.year],
            months=[month_to_process.month],
            days=list(range(1, 32)),
            target_location=download_file,
        )

        format_downloaded_file(download_file, allowed_factors)
        download_file.rename(file_path.with_suffix(_FORMATTED_SUFFIX))
        finalize_formatted_file(file_path, month_to_process, most_recent_data_date)


def file_requires_update(file_path: Path, current_month: date, verification_date: date):
    if file_path.with_suffix(".nc").exists():
        # A regular file exists, no updates required
        logger.debug("A regular file already exists: NO UPDATE REQUIRED")
        return False

    if file_path.with_suffix(_TEMP_SUFFIX).exists():
        # If a file is temporary we only check for a permanent update if more than 3 months have past since the current
        # most recent date with data.
        threshold_date = (verification_date - relativedelta(months=3)).replace(day=1)
        if current_month < threshold_date:
            logger.debug("A temporary file exists within the update range: UPDATE REQUIRED")
            return True
        logger.debug("A temporary file exists within the update range: UPDATE REQUIRED")
        return False

    # A file exists but isn't any regular supported type to be updated
    if file_path.with_suffix(_UNFORMATTED_SUFFIX).exists() or file_path.with_suffix(_FORMATTED_SUFFIX).exists():
        logger.debug("An unformatted file or formatted file exists: UPDATE REQUIRED")
        return True  # An update should both clean the UNFORMATTED file and generate a proper one

    if not file_path.with_suffix(".nc").exists() or file_path.with_suffix(_INCOMPLETE_SUFFIX).exists():
        logger.debug("No file exists, or it is still incomplete: UPDATE REQUIRED ")
        return True  # No file matching the mask or incomplete files always mean the update is required!

    files_in_folder = glob.glob(f"{file_path}*.nc")
    logger.warning(f"Unexpected files existed in the repository folder: {files_in_folder}. These should be dealt with.")
    return False


def format_downloaded_file(unformatted_file: Path, allowed_factors: Dict):
    logger.info(f"Formatting the downloaded file at: {unformatted_file}")
    ds_unformatted = load_file(unformatted_file)
    ds_unformatted.attrs = {}  # Remove unneeded attributes

    if "expver" in ds_unformatted.indexes.keys():
        # We remove the expver index used to denominate temporary data (5) and regular data (1) and add a field for it
        # NOTE: We removed the drop_sel version as it didn't quite have the same result as drop yet. Reverting until
        #  the proper use has been validated...
        ds_unformatted_expver5 = ds_unformatted.sel(expver=5).drop("expver").dropna("time", how="all")
        ds_unformatted_expver1 = ds_unformatted.sel(expver=1).drop("expver").dropna("time", how="all")

        # Recombine the data
        ds_unformatted = ds_unformatted_expver1.merge(ds_unformatted_expver5)
        ds_unformatted["is_permanent_data"] = False
    else:
        ds_unformatted["is_permanent_data"] = True

    # Rename the factors to their longer names:
    for factor in ds_unformatted.variables.keys():
        if factor in allowed_factors:
            ds_unformatted = ds_unformatted.rename_vars({factor: allowed_factors[factor]})

    # Rename and encode data where needed:
    ds_unformatted.time.encoding["units"] = "hours since 2016-01-01"
    ds_unformatted = ds_unformatted.rename(name_dict={"latitude": "lat", "longitude": "lon"})

    # Store the data
    ds_unformatted.to_netcdf(path=unformatted_file, format="NETCDF4", engine="netcdf4")


def finalize_formatted_file(file_path: Path, current_month: date, verification_date: date):
    logger.info(f"Finalizing the formatted file for location: {file_path}")
    formatted_file = file_path.with_suffix(_FORMATTED_SUFFIX)
    incomplete_month = verification_date.replace(day=1)
    permanent_month = (verification_date - relativedelta(months=3)).replace(day=1)

    if not formatted_file.exists():
        logger.error(f"Unexpected error: Expected file [{formatted_file}] did not exist!")
        raise FileNotFoundError(f"The pre-formatted month file [{formatted_file}] was not found!")

    # Cleanup old files:
    for file_suffix in [_TEMP_SUFFIX, _INCOMPLETE_SUFFIX]:
        if file_path.with_suffix(file_suffix).exists():
            _safely_delete_file(file_path.with_suffix(file_suffix))

    # Rename the file to its proper name:
    if current_month == verification_date.replace(day=1):
        # Current month means an incomplete file
        file_path.with_suffix(_FORMATTED_SUFFIX).rename(file_path.with_suffix(_INCOMPLETE_SUFFIX))
        logger.debug(f"Month [{current_month}] was renamed to: {file_path.with_suffix(_INCOMPLETE_SUFFIX)}")
    elif permanent_month < current_month < incomplete_month:
        # Non-permanent file
        file_path.with_suffix(_FORMATTED_SUFFIX).rename(file_path.with_suffix(_TEMP_SUFFIX))
        logger.debug(f"Month [{current_month}] was renamed to: {file_path.with_suffix(_TEMP_SUFFIX)}")
    else:
        # Permanent file
        file_path.with_suffix(_FORMATTED_SUFFIX).rename(file_path.with_suffix(".nc"))
        logger.debug(f'Month [{current_month}] was renamed to: {file_path.with_suffix(".nc")}')


def _get_actual_most_recent_data_for_model(
    date_to_start_from: date, era5_dataset_name: str, era5_product_type: str, grid_resolution: Tuple[float, float]
):
    got_proper_response = False
    date_to_check = date_to_start_from
    while not got_proper_response:
        try:
            download_era5_file(
                dataset=era5_dataset_name,
                product_type=era5_product_type,
                grid_resolution=grid_resolution,
                weather_factors=[
                    "stl1",
                ],  # Only one factor that exists in both datasets
                years=[date_to_check.year],
                months=[date_to_check.month],
                days=[date_to_check.day],
                target_location=tempfile.NamedTemporaryFile().name,
            )
            got_proper_response = True
        finally:
            if date_to_check < date_to_start_from - relativedelta(years=1):
                raise IOError("No valid starting date could be found within more than a year time..")
            date_to_check = date_to_check - relativedelta(days=1)

    logger.info(f"Determined the first valid processable day to be: {date_to_check}")
    return date_to_check


def download_era5_file(
    dataset,
    product_type,
    grid_resolution,
    weather_factors,
    years,
    months,
    days,
    target_location,
):
    """A function that download a NetCDF file to the target location, containing the requested factors for an
        also requested range of years, months, days and locations.

    Args:
        dataset:            The name of the dataset to download from
        product_type:       The type of product to download
        grid_resolution:    The grid resolution to download the results in
        weather_factors:    A list of factors to be downloaded (in string format)
        years:              A list of years to be requested (numeric)
        months:             A list of months to be requested (numeric)
        days:               A list of days to be requested (numeric)
        target_location:    The file location to which the result should be saved after downloading.

    Returns:
        Returns nothing, but success means a result file will exist at the given target location.
    """
    c = downloader.CDSDownloadClient(persist_request_callback=_repository_callback(), verify=True)
    try:
        c.retrieve(
            dataset,
            {
                "product_type": product_type,
                "variable": weather_factors,
                "year": years,
                "month": months,
                "day": days,
                "time": [
                    "00:00",
                    "01:00",
                    "02:00",
                    "03:00",
                    "04:00",
                    "05:00",
                    "06:00",
                    "07:00",
                    "08:00",
                    "09:00",
                    "10:00",
                    "11:00",
                    "12:00",
                    "13:00",
                    "14:00",
                    "15:00",
                    "16:00",
                    "17:00",
                    "18:00",
                    "19:00",
                    "20:00",
                    "21:00",
                    "22:00",
                    "23:00",
                ],
                "area": [50.75, 3.2, 53.7, 7.22],
                "grid": [grid_resolution[0], grid_resolution[1]],
                "format": "netcdf",
            },
            target_location,
        )
    except Exception as e:
        logger.warning(f"Could not download the requested file: {e}")
        return False


def load_file(file: Path) -> xr.Dataset:
    """A function that loads and returns the full data for a specific repository file as a Xarray Dataset
    Args:
        file:   The filename (in the Path format by PathLib) specifying the file to load
    Returns:
        An Xarray Dataset containing all the weather data held within the specified file.

    """
    if file.exists():
        with xr.open_dataset(file) as ds:
            ds.load()
        return ds

    # Raise a FileNotFoundError if the file doesn't exist
    logger.error(f"File [{file!s}] does not exist")
    raise FileNotFoundError


def _safely_delete_file(file: Path):
    """Basic function to safely remove files from the repository if possible, and supply errors if not"""
    try:
        logger.debug(f"Safely deleting file [{file}]")
        file.unlink()
    except OSError as e:
        logger.error(f"Could not safely delete file: {e}")
        raise OSError(f"Could not safely delete file: {file}")
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta
from typing import Optional, Union

import numpy as np
import pandas as pd
from fastapi import HTTPException
from loguru import logger


def parse_datetime(
    datetime_string,
    round_missing_time_up=False,
    round_to_days=False,
    raise_errors=False,
    loc=None,
) -> Optional[datetime]:
    if datetime_string is None:
        return None

    dt = pd.to_datetime(datetime_string, dayfirst=False, errors="coerce")

    if pd.isnull(dt):
        logger.exception("Error while parsing datetime string", input=datetime_string)
        if raise_errors:
            # Note: replace when FastAPI supports Pydantic models to define query parameters
            # (meaning Validators can be used)
            error_msg = {
                "loc": loc,
                "msg": "invalid datetime format",
                "type": "type_error.datetime",
            }
            raise HTTPException(status_code=422, detail=[error_msg])

        dt = None

    if dt is not None and (round_missing_time_up or round_to_days) and time_unknown(dt, datetime_string):
        if round_to_days:
            dt = dt + timedelta(days=1)
        else:
            dt = dt.replace(hour=23, minute=59, second=59)

    if dt is not None:
        dt = np.datetime64(dt).astype(datetime)

    return dt


def time_unknown(dt: datetime, datetime_string: str):  # pragma: no cover
    if dt.hour == 0 and dt.minute == 0 and dt.second == 0 and ":" not in datetime_string:
        return True
    return False


def validate_begin_and_end(
    start: datetime, end: datetime, data_start: Union[datetime, None] = None, data_end: Union[datetime, None] = None
):
    """Checks the given date parameters and replaces them with default values if they aren't valid.
    The resulting values are then returned.
    """
    if data_end is None:
        # Assuming predictions fill in this value, the most recent value for the past is before "now".
        data_end = datetime.utcnow()

    if data_start is not None and data_start > start:
        # If the starting moment lies before what can be requested, put it at the moment from which it can be requested
        start = data_start

    if data_end is not None and data_end < end:
        end = data_end

    if start >= data_end:
        raise HTTPException(
            422, f"Invalid [start] value [{start}]: value lies after last available moment for model ({data_end})"
        )
    if data_start is not None and end <= data_start:
        raise HTTPException(
            422, f"Invalid [end] value [{end}]: value lies before first available moment for model ({data_start})"
        )

    if end < start:
        raise HTTPException(422, f"Invalid [start] and [end] values: [end]({end}) lies before [start]({start})")

    return start, end



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#
# 2019: Modified version by Alliander to improve scalability.

import json
import os
import time
from pathlib import Path

import requests
from loguru import logger


def bytes_to_string(n):
    u = ["", "K", "M", "G", "T", "P"]
    i = 0
    while n >= 1024:
        n /= 1024.0
        i += 1
    return "%g%s" % (int(n * 10 + 0.5) / 10.0, u[i])


def read_rc_file(path: Path):
    config = {}
    with open(path) as f:
        for line in f.readlines():
            if ":" in line:
                k, v = line.strip().split(":", 1)
                if k in ("url", "key", "verify"):
                    config[k] = v.strip()
    return config


class Result(object):
    def __init__(self, client, reply):
        self.reply = reply
        self._url = client.api_url
        self.session = client.session
        self.robust = client.robust
        self.verify = client.api_verify
        self.cleanup = client.delete
        self._deleted = False

    def _download(self, url: str, size, target):
        if not target:
            target = url.split("/")[-1]

        logger.info(f"Downloading [{url}] to [{target}]. ({bytes_to_string(size)})")
        start = time.time()

        r = self.robust(requests.get)(url, stream=True, verify=self.verify)
        try:
            r.raise_for_status()
            total = 0

            with open(target, "wb") as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
                        total += len(chunk)
        finally:
            r.close()

        if total != size:
            raise EOFError(f"Download failed due to incorrect file size: Downloaded {total} byte(s) out of {size}")

        elapsed = time.time() - start
        if elapsed:
            logger.info(f"Download rate: {bytes_to_string(size / elapsed)} bytes per second")
        return target

    def download(self, target=None):
        return self._download(self.location, self.content_length, target)

    @property
    def content_length(self):
        return int(self.reply["content_length"])

    @property
    def location(self):
        return self.reply["location"]

    @property
    def content_type(self):
        return self.reply["content_type"]

    def __repr__(self):
        return (
            f"Result(content_length={self.content_length}, content_type={self.content_type}, "
            f"location={self.location})"
        )

    def check(self):
        logger.debug(f'HEAD {self.reply["location"]}')
        metadata = self.robust(self.session.head)(self.reply["location"], verify=self.verify)
        metadata.raise_for_status()

    def delete(self):
        if self._deleted:
            return

        if "request_id" in self.reply:
            rid = self.reply("request_id")
            task_url = f"{self._url}/tasks/{rid}"

            delete = self.session.delete(task_url, verify=self.verify)
            logger.debug(f"DELETE returns: ({delete.status_code}) {delete.reason}")

            try:
                delete.raise_for_status()
            except RuntimeError:
                logger.warning(f"DELETE [{task_url}] returns: ({delete.status_code}) {delete.reason}")

            self._deleted = True

    def __del__(self):
        try:
            if self.cleanup:
                self.delete()
        except Exception as e:
            logger.error(e)
            raise e


class CDSDownloadClient(object):
    def __init__(
        self,
        url: str = os.environ.get("CDSAPI_URL"),
        key: str = os.environ.get("CDSAPI_KEY"),
        verify=None,
        timeout=None,
        full_stack: bool = False,
        delete: bool = False,
        retry_max: int = 500,
        sleep_max: int = 120,
        info_callback=None,
        warning_callback=None,
        error_callback=None,
        debug_callback=None,
        persist_request_callback=None,
    ):
        self.api_url, self.api_key, self.api_verify = self._load_cdsapi_config(url, key, verify)

        self.timeout = timeout
        self.sleep_max = sleep_max
        self.retry_max = retry_max
        self.full_stack = full_stack
        self.delete = delete
        self.last_state = None

        # Callbacks:
        self.info_callback = info_callback
        self.debug_callback = debug_callback
        self.warning_callback = warning_callback
        self.error_callback = error_callback
        self.persist_request_callback = persist_request_callback

        #
        self.persisted_request = False
        self.session = requests.Session()
        self.session.auth = tuple(self.api_key.split(":", 2))

        settings_dict = dict(
            url=self.api_url,
            key=self.api_key,
            verify=self.api_verify,
            timeout=self.timeout,
            sleep_max=self.sleep_max,
            retry_max=self.retry_max,
            full_stack=self.full_stack,
            delete=self.delete,
        )

        logger.debug(f"CDSAPI Downloader Settings: {settings_dict}")

    def retrieve(self, name, request, target=None, request_id=None):
        """This function retrieves a download-result and stores it at the target location if given."""
        logger.info(f"Starting retrieval of URL:{self.api_url}/resources/{name}")
        result = self._request_retrieval(
            f"{self.api_url}/resources/{name}", request, request_id if request_id else None
        )
        if target:
            result.download(target)
        return result

    def _request_retrieval(self, url, request, request_id=None):
        reply = self._request_handler(url, request, request_id)

        sleep = 1
        start = time.time()

        while True:
            logger.debug(f"REPLY: {reply}")

            reply_state = reply["state"]

            if reply_state != self.last_state:
                logger.info(f"Request is: {reply_state}")
                self.last_state = reply_state

            if reply_state == "completed":
                logger.debug("Request completed!")
                return Result(self, reply)

            if reply_state in ("queued", "running", "resuming_download"):
                sleep, reply = self._active_request_handler(reply, url, request, request_id, start, sleep)
                continue

            if reply_state == "failed":
                logger.error(f"An error occurred: {reply['error'].get('message')}")
                logger.error(f"The following reason was given: {reply['error'].get('reason')}")

                for part in reply.get("error", {}).get("context", {}).get("traceback", "").split("\n"):
                    if part.strip() == "" and not self.full_stack:
                        break
                    logger.error(f" {part}")

                raise RuntimeError(f'{reply["error"].get("message")} | {reply["error"].get("reason")}')

            raise RuntimeError(f"UNKNOWN API STATE: [{reply_state}]")

    def _request_handler(self, url, request, request_id):
        if request_id is None:
            logger.info(f"Sending request to {url}")
            logger.debug(f"POST {url} {json.dumps(request)}")

            result = self.robust(self.session.post)(url, json=request, verify=self.api_verify)
            reply = None

            try:
                result.raise_for_status()
                reply = result.json()
            except RuntimeError:
                if reply is None:
                    try:
                        reply = result.json()
                    except RuntimeError:
                        reply = dict(message=result.text)
                logger.debug(json.dumps(reply))

                if "message" in reply:
                    error = reply["message"]

                    if "context" in reply and "required_terms" in reply["context"]:
                        e = [error]
                        for term in reply["context"]["required_terms"]:
                            e.append(
                                f"To access this resource, you'll first need to accept the terms of {term['title']} "
                                f"at: {term['url']}"
                            )
                            error = ". ".join(e)
                    raise RuntimeError(error)
                else:
                    raise RuntimeError("An unknown error occurred...")
        else:
            logger.info("Resuming the download task and skipping initial request to API.")
            reply = {"state": "resuming_download", "request_id": request_id}

        return reply

    def _active_request_handler(self, reply, url, request, request_id, start, sleep):
        rid = reply["request_id"]  # Other Request ID?

        if self.persist_request_callback and not self.persisted_request:
            self.persist_request_callback(url=url, request=request, request_id=request_id, start_timestamp=start)
            self.persisted_request = True

        if self.timeout and (time.time() - start > self.timeout):
            raise TimeoutError("Request Timed Out!")

        logger.debug(f"The Request ID is {rid} (sleeping for {sleep} seconds)")
        time.sleep(sleep)
        sleep *= 1.5

        sleep = self.sleep_max if sleep > self.sleep_max else sleep

        task_url = f"{self.api_url}/tasks/{rid}"
        logger.debug(f"GET: {task_url}")

        result = self.robust(self.session.get)(task_url, verify=self.api_verify)
        result.raise_for_status()
        reply = result.json()
        return sleep, reply

    @staticmethod
    def _load_cdsapi_config(url, key, verify):
        dotrc_file = Path(os.environ.get("CDSAPI_RC", str(Path.home().joinpath(".cdsapirc"))))
        cdsapi_key = os.environ.get("CDSAPI_KEY", None)
        cdsapi_url = os.environ.get("CDSAPI_URL", None)

        # Prefer the CDSAPI RC file over CDSAPI_KEY and CDSAPI_URL:
        if (not url or not key) and dotrc_file.exists():
            logger.debug("CDS API [.rc]-file exists. ")
            config = read_rc_file(dotrc_file)

            if not key:
                key = config.get("key")
            if not url:
                url = config.get("url")
            if not verify:
                verify = int(config.get("verify", 1))

        # If key or url are still missing try using CDSAPI_KEY and/or CDSAPI_URL:
        if (not url or not key) and cdsapi_key and cdsapi_url:
            key = cdsapi_key
            url = cdsapi_url

        # If key or url are still missing, or no verify status was passed, exit with error:
        if not url or not key or not verify:
            raise AttributeError(
                f"Missing or incomplete configuration. "
                f"Please create or verify the file at [{dotrc_file}], or create/verify "
                f"the CDSAPI_KEY and CDSAPI_URL environment variables."
            )
        logger.info("Successfully loaded settings for the CDS Downloader!")

        return url, key, True if verify else False

    def robust(self, call):
        def retryable(code, reason):
            if code in [
                requests.codes.internal_server_error,
                requests.codes.bad_gateway,
                requests.codes.service_unavailable,
                requests.codes.gateway_timeout,
                requests.codes.too_many_requests,
                requests.codes.request_timeout,
            ]:
                return True
            logger.debug(f"The current response code is considered not retryable: ({code}) {reason}")
            return False

        def wrapped(*args, **kwargs):

            tries = 0
            while tries < self.retry_max:
                logger.info(f"WRAPPED [Tries: {tries} ({self.retry_max})]")
                try:
                    r = call(*args, **kwargs)
                except requests.exceptions.ConnectionError as e:
                    r = None
                    logger.warning(f"Recovering from ConnectionError [{e}]. (Attempt #{tries}) of {self.retry_max})")
                if r:
                    if not retryable(r.status_code, r.reason):
                        return r
                    logger.warning(
                        f"Recovering from HTTPError [{r.status_code, r.reason}]. "
                        f"(Attempt #{tries} of {self.retry_max})"
                    )

                tries += 1
                logger.warning(f"Retrying after {self.sleep_max} seconds..")
                time.sleep(self.sleep_max)

        return wrapped



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import glob
from datetime import datetime
from typing import List

from dateutil.relativedelta import relativedelta
from loguru import logger

from weather_provider_api.routers.weather.repository.repository import (
    WeatherRepositoryBase,
)
from weather_provider_api.routers.weather.sources.cds.client.utils_era5 import (
    era5_update,
)
from weather_provider_api.routers.weather.sources.cds.factors import era5land_factors
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.grid_helpers import (
    round_coordinates_to_wgs84_grid,
)


class ERA5LandRepository(WeatherRepositoryBase):
    """A class that holds all functionality (excepting the downloader) for the ERA5 Single Levels Repository
    """

    def __init__(self):
        super().__init__()
        self.repository_name = "CSD: ERA5-Land"
        logger.debug(f"Initializing {self.repository_name} repository")
        self.file_prefix = "ERA5LAND"
        self.runtime_limit = 3 * 60  # 3 hours maximum runtime
        self.permanent_suffixes = ["INCOMPLETE", "TEMP"]
        self.grid_resolution = 0.25
        self.file_identifier_length = 7
        self.age_of_permanence_in_months = 3

        logger.debug(f"Initialized {self.repository_name} repository")

    @staticmethod
    def _get_repo_sub_folder():
        return "ERA5LAND"

    @property
    def first_day_of_repo(self):
        first_day_of_repo = datetime.utcnow() - relativedelta(years=12, days=5)
        first_day_of_repo = first_day_of_repo.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        return first_day_of_repo

    @property
    def last_day_of_repo(self):
        last_day_of_repo = datetime.utcnow() - relativedelta(days=2)
        last_day_of_repo = last_day_of_repo.replace(hour=0, minute=0, second=0, microsecond=0)
        return last_day_of_repo

    def update(self):
        """The implementation of the WeatherRepository required update() function.
            This function handles all the required actions to update the repository completely, but taking into
            account its set runtime_limit. If based on the time of completion of other downloaded files this session
            the next file wouldn't complete within the runtime_limit, the update process halts.
            (if no other downloads were made yet, a generous rough estimate is used)

        Returns:
            A RepositoryUpdateResult value indicating a completion, time-out or failure of the update process
        """
        # Always start with a nicely cleaned repository
        self.cleanup()

        return era5_update(
            self.file_prefix,
            self.repository_folder,
            (self.first_day_of_repo, self.last_day_of_repo),
            "reanalysis-era5-land",
            "reanalysis",
            [era5land_factors[x] for x in list(era5land_factors.keys())],
            (self.grid_resolution, self.grid_resolution),
            era5land_factors,
            self.runtime_limit,
            True,
        )

    def _delete_files_outside_of_scope(self):
        """A function that deletes all files in the repository with a date not inside the repository's scope.
            All files labeled as either before or after the given scope will be deleted.

        Returns:
            Nothing. Successful means the all files outside the scope were deleted.
        """
        len_filename_until_date = len(str(self.repository_folder.joinpath(self.file_prefix))) + 1

        for file_name in glob.glob(str(self.repository_folder.joinpath(self.file_prefix)) + "*.nc"):
            file_year = int(file_name[len_filename_until_date : len_filename_until_date + 4])
            file_month = int(file_name[len_filename_until_date + 5 : len_filename_until_date + 7])

            if (
                file_year < self.first_day_of_repo.year
                or file_year > self.last_day_of_repo.year
                or (file_year == self.first_day_of_repo.year and file_month < self.first_day_of_repo.month)
                or (file_year == self.last_day_of_repo.year and file_month > self.last_day_of_repo.month)
            ):
                logger.debug(
                    f"Deleting file [{file_name}] because it does not lie in the "
                    f"repository scope ({self.first_day_of_repo, self.last_day_of_repo})"
                )
                self._safely_delete_file(file_name)

    def _get_file_list_for_period(self, start: datetime, end: datetime):
        """A function that retrieves a list of files in the repository associated with the requested period of time
        Args:
            start:  A datetime containing the start of the requested period of time.
            end:    A datetime containing the end of the requested period of time.

        Returns:
            A list of files (in string format) that indicate the files containing data for the requested period.
        """
        self.cleanup()

        len_filename_until_date = len(str(self.repository_folder.joinpath(self.file_prefix))) + 1
        full_list_of_files = glob.glob(str(self.repository_folder.joinpath(self.file_prefix)) + "*.nc")
        list_of_filtered_files = []
        for file in full_list_of_files:
            file_year = int(file[len_filename_until_date : len_filename_until_date + 4])
            file_month = int(file[len_filename_until_date + 5 : len_filename_until_date + 7])
            date_for_filename = datetime(year=file_year, month=file_month, day=15)

            if start.replace(day=1) < date_for_filename < datetime(year=end.year, month=end.month, day=28):
                # If the file is within the requested period, save it to the list of filtered files
                list_of_filtered_files.append(file)

        return list_of_filtered_files

    def get_grid_coordinates(self, coordinates: List[GeoPosition]) -> List[GeoPosition]:
        # Rounds a list of GeoPositions to the resolution set through grid_resolution
        return round_coordinates_to_wgs84_grid(coordinates, (self.grid_resolution, self.grid_resolution))
